@@ -1,29 +1,31 @@
+/*
+     PLIB - A Suite of Portable Game Libraries
+     Copyright (C) 1998,2002  Steve Baker
+ 
+     This library is free software; you can redistribute it and/or
+     modify it under the terms of the GNU Library General Public
+     License as published by the Free Software Foundation; either
+     version 2 of the License, or (at your option) any later version.
+ 
+     This library is distributed in the hope that it will be useful,
+     but WITHOUT ANY WARRANTY; without even the implied warranty of
+     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+     Library General Public License for more details.
+ 
+     You should have received a copy of the GNU Library General Public
+     License along with this library; if not, write to the Free Software
+     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
+ 
+     For further information visit http://plib.sourceforge.net
+
+     $Id: sg.h,v 1.47 2002/09/02 06:05:46 sjbaker Exp $
+*/
+
 
 #ifndef SG_H
 #define SG_H  1
 
-#include <limits.h>
-#include <math.h>
-
-/* the next lines are to define BSD */
-/* see http://www.freebsd.org/handbook/porting.html for why we do this */
-
-#if (defined(__unix__) || defined(unix)) && !defined(USG)
-#include <sys/param.h>
-#endif
- 
-#ifdef BSD
-#include <float.h>
-#endif
-#ifdef __MWERKS__
-#include <float.h>
-#endif
-#ifdef WIN32
-#include <float.h>
-#endif
-#ifdef __CYGWIN__
-#include <float.h>
-#endif
+#include "ul.h"
 
 #define sgFloat float
 #define SGfloat float
@@ -34,6 +36,8 @@
 #define SG_TWO   2.0f
 #define SG_THREE 3.0f
 #define SG_45    45.0f
+#define SG_60    60.0f
+#define SG_90    90.0f
 #define SG_180   180.0f
 #define SG_MAX   FLT_MAX
 
@@ -58,7 +62,27 @@
    
 inline SGfloat sgSqrt   ( const SGfloat x ) { return (SGfloat) sqrt ( x ) ; }
 inline SGfloat sgSquare ( const SGfloat x ) { return x * x ; }
-inline SGfloat sgAbs    ( const SGfloat a ) { return ( a < SG_ZERO ) ? -a : a ; }
+inline SGfloat sgAbs    ( const SGfloat a ) { return (a<SG_ZERO) ? -a : a ; }
+
+/* 
+  Type-casted sin/cos/tan/asin/acos/atan2 ANGLES IN DEGREES
+*/
+
+inline SGfloat sgASin ( SGfloat s )
+                { return (SGfloat) asin ( s ) * SG_RADIANS_TO_DEGREES ; }
+inline SGfloat sgACos ( SGfloat s )
+                { return (SGfloat) acos ( s ) * SG_RADIANS_TO_DEGREES ; }
+inline SGfloat sgATan ( SGfloat s )
+                { return (SGfloat) atan ( s ) * SG_RADIANS_TO_DEGREES ; }
+inline SGfloat sgATan2 ( SGfloat y, SGfloat x )
+                { return (SGfloat) atan2 ( y,x ) * SG_RADIANS_TO_DEGREES ; }
+inline SGfloat sgSin ( SGfloat s )
+                { return (SGfloat) sin ( s * SG_DEGREES_TO_RADIANS ) ; }
+inline SGfloat sgCos ( SGfloat s )
+                { return (SGfloat) cos ( s * SG_DEGREES_TO_RADIANS ) ; }
+inline SGfloat sgTan ( SGfloat s )
+                { return (SGfloat) tan ( s * SG_DEGREES_TO_RADIANS ) ; }
+
 
 inline int sgCompareFloat ( const SGfloat a, const SGfloat b, const SGfloat tol )
 {
@@ -128,6 +152,9 @@ inline void sgMakeCoordMat4( sgMat4 dst, const sgCoord *src )
   sgMakeCoordMat4 ( dst, src->xyz, src->hpr ) ;
 }
 
+extern void sgMakeLookAtMat4 ( sgMat4 dst,
+        const sgVec3 eye, const sgVec3 center, const sgVec3 up ) ;
+
 extern void sgMakeRotMat4   ( sgMat4 dst, const SGfloat angle, const sgVec3 axis ) ;
 
 inline void sgMakeRotMat4   ( sgMat4 dst, const sgVec3 hpr )
@@ -169,6 +196,26 @@ inline void sgXformPnt4     ( sgVec4 dst, const sgMat4 mat ) { sgXformPnt4 ( dst
 inline void sgFullXformPnt3 ( sgVec3 dst, const sgMat4 mat ) { sgFullXformPnt3 ( dst, dst, mat ) ; }
 inline void sgFullXformPnt4 ( sgVec4 dst, const sgMat4 mat ) { sgFullXformPnt4 ( dst, dst, mat ) ; }
 
+
+/* Bits returned by sgClassifyMat4 */
+
+#define SG_IDENTITY        0x00   // for clarity
+#define SG_ROTATION        0x01   // includes a rotational component
+#define SG_MIRROR          0x02   // changes handedness (det < 0)
+#define SG_SCALE           0x04   // uniform scaling
+#define SG_NONORTHO        0x10   // 3x3 not orthogonal
+#define SG_TRANSLATION     0x20   // translates
+#define SG_PROJECTION      0x40   // forth column not 0,0,0,1
+
+/* Are these needed? sgClassifyMat4() does set the general scale bit for some matrices,
+ * but it is not easily defined. Use SG_NONORTHO instead (which is also set). */
+#define SG_UNIFORM_SCALE   SG_SCALE
+#define SG_GENERAL_SCALE   0x08   // x, y and z scaled differently
+
+extern int sgClassifyMat4 ( const sgMat4 mat ) ;
+
+
+
 /*
   Basic low-level vector functions.
   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
@@ -461,35 +508,29 @@ inline void sgAddScaledVec4 ( sgVec4 dst, const sgVec4 src1, const sgVec4 src2,
 
 inline int sgCompareVec2 ( const sgVec2 a, const sgVec2 b, const SGfloat tol )
 {
-  int val = 0 ;
+  if ( sgCompareFloat( a[0], b[0], tol ) != 0 ) return FALSE ;
+  if ( sgCompareFloat( a[1], b[1], tol ) != 0 ) return FALSE ;
 
-  if ( ( val = sgCompareFloat( a[0], b[0], tol ) ) != 0 ) return val ;
-  if ( ( val = sgCompareFloat( a[1], b[1], tol ) ) != 0 ) return val ;
-
-  return 0 ;
+  return TRUE ;
 }
 
 inline int sgCompareVec3 ( const sgVec3 a, const sgVec3 b, const SGfloat tol )
 {
-  int val = 0 ;
+  if ( sgCompareFloat( a[0], b[0], tol ) != 0 ) return FALSE ;
+  if ( sgCompareFloat( a[1], b[1], tol ) != 0 ) return FALSE ;
+  if ( sgCompareFloat( a[2], b[2], tol ) != 0 ) return FALSE ;
 
-  if ( ( val = sgCompareFloat( a[0], b[0], tol ) ) != 0 ) return val ;
-  if ( ( val = sgCompareFloat( a[1], b[1], tol ) ) != 0 ) return val ;
-  if ( ( val = sgCompareFloat( a[2], b[2], tol ) ) != 0 ) return val ;
-
-  return 0 ;
+  return TRUE ;
 }
 
 inline int sgCompareVec4 ( const sgVec4 a, const sgVec4 b, const SGfloat tol )
 {
-  int val = 0 ;
+  if ( sgCompareFloat( a[0], b[0], tol ) != 0 ) return FALSE ;
+  if ( sgCompareFloat( a[1], b[1], tol ) != 0 ) return FALSE ;
+  if ( sgCompareFloat( a[2], b[2], tol ) != 0 ) return FALSE ;
+  if ( sgCompareFloat( a[3], b[3], tol ) != 0 ) return FALSE ;
 
-  if ( ( val = sgCompareFloat( a[0], b[0], tol ) ) != 0 ) return val ;
-  if ( ( val = sgCompareFloat( a[1], b[1], tol ) ) != 0 ) return val ;
-  if ( ( val = sgCompareFloat( a[2], b[2], tol ) ) != 0 ) return val ;
-  if ( ( val = sgCompareFloat( a[3], b[3], tol ) ) != 0 ) return val ;
-
-  return 0 ;
+  return TRUE ;
 }
 
 
@@ -533,6 +574,56 @@ inline SGfloat sgScalarProductVec4 ( const sgVec4 a, const sgVec4 b )
 
 extern void sgVectorProductVec3 ( sgVec3 dst, const sgVec3 a, const sgVec3 b ) ;
 
+inline SGfloat sgLerp ( const SGfloat a, const SGfloat b, const SGfloat f )
+{
+  return a + f * ( b - a ) ;
+}
+
+inline void sgLerpVec4 ( sgVec4 dst, const sgVec4 a, const sgVec4 b, const SGfloat f )
+{
+  dst[0] = a[0] + f * ( b[0] - a[0] ) ;
+  dst[1] = a[1] + f * ( b[1] - a[1] ) ;
+  dst[2] = a[2] + f * ( b[2] - a[2] ) ;
+  dst[3] = a[3] + f * ( b[3] - a[3] ) ;
+}
+
+
+inline void sgLerpVec3 ( sgVec3 dst, const sgVec3 a, const sgVec3 b, const SGfloat f )
+{
+  dst[0] = a[0] + f * ( b[0] - a[0] ) ;
+  dst[1] = a[1] + f * ( b[1] - a[1] ) ;
+  dst[2] = a[2] + f * ( b[2] - a[2] ) ;
+}
+
+
+inline void sgLerpVec2 ( sgVec2 dst, const sgVec2 a, const sgVec2 b, const SGfloat f )
+{
+  dst[0] = a[0] + f * ( b[0] - a[0] ) ;
+  dst[1] = a[1] + f * ( b[1] - a[1] ) ;
+}
+
+
+inline void sgLerpAnglesVec3 ( sgVec3 dst, const sgVec3 a,
+                                           const sgVec3 b,
+                                           const SGfloat f )
+{
+  sgVec3 tmp ;
+ 
+  if ( b[0] - a[0] >  180.0f ) tmp[0] = a[0] + 360.0f ; else
+  if ( b[0] - a[0] < -180.0f ) tmp[0] = a[0] - 360.0f ; else tmp[0] = a[0] ;
+ 
+  if ( b[1] - a[1] >  180.0f ) tmp[1] = a[1] + 360.0f ; else
+  if ( b[1] - a[1] < -180.0f ) tmp[1] = a[1] - 360.0f ; else tmp[1] = a[1] ;
+ 
+  if ( b[2] - a[2] >  180.0f ) tmp[2] = a[2] + 360.0f ; else
+  if ( b[2] - a[2] < -180.0f ) tmp[2] = a[2] - 360.0f ; else tmp[2] = a[2] ;
+ 
+  dst[0] = tmp[0] + f * ( b[0] - tmp[0] ) ;
+  dst[1] = tmp[1] + f * ( b[1] - tmp[1] ) ;
+  dst[2] = tmp[2] + f * ( b[2] - tmp[2] ) ;
+}                                                                               
+
+
 
 inline SGfloat sgDistanceSquaredVec2 ( const sgVec2 a, const sgVec2 b )
 {
@@ -690,26 +781,69 @@ inline void sgMakeIdentMat4 ( sgMat4 dst )
   sgSetVec4 ( dst[3], SG_ZERO, SG_ZERO, SG_ZERO, SG_ONE  ) ;
 }
 
+
+extern void sgMakePickMatrix( sgMat4 mat, sgFloat x, sgFloat y,
+                    sgFloat width, sgFloat height, sgVec4 viewport ) ;
+
 extern int  sgCompare3DSqdDist ( const sgVec3 a, const sgVec3 b, const SGfloat sqd_dist ) ;
-extern void sgMakeTransMat4 ( sgMat4 m, const SGfloat x, const SGfloat y, const SGfloat z ) ;
-extern void sgMakeTransMat4 ( sgMat4 m, const sgVec3 xyz ) ;
-extern void sgMakeCoordMat4 ( sgMat4 m, const SGfloat x, const SGfloat y, const SGfloat z,
-                                        const SGfloat h, const SGfloat p, const SGfloat r ) ;
-extern void sgMakeCoordMat4 ( sgMat4 m, const sgCoord *c ) ;
 
 inline SGfloat sgDistToLineVec2 ( const sgVec3 line, const sgVec2 pnt )
 {
   return sgScalarProductVec2 ( line, pnt ) + line[2] ;
 }
  
+
+struct sgLineSegment3   /* Bounded line segment */
+{
+  sgVec3 a ;
+  sgVec3 b ;
+} ;
+
+struct sgLine3    /* Infinite line */
+{
+  sgVec3 point_on_line ;
+  sgVec3 direction_vector ;  /* Should be a unit vector */
+} ;
+
+
+inline void sgLineSegment3ToLine3 ( sgLine3 *line,
+                                   const sgLineSegment3 lineseg )
+{
+  sgCopyVec3      ( line->point_on_line   , lineseg.a ) ;
+  sgSubVec3       ( line->direction_vector, lineseg.b, lineseg.a ) ;
+  sgNormaliseVec3 ( line->direction_vector ) ;
+}
+
+
+SGfloat sgDistSquaredToLineVec3        ( const sgLine3 line,
+                                         const sgVec3 pnt ) ;
+SGfloat sgDistSquaredToLineSegmentVec3 ( const sgLineSegment3 line,
+                                         const sgVec3 pnt ) ;
+
+
+inline SGfloat sgDistToLineVec3 ( const sgLine3 line,
+                                  const sgVec3 pnt )
+{
+  return sgSqrt ( sgDistSquaredToLineVec3 ( line, pnt ) );
+}
+
+
+inline SGfloat sgDistToLineSegmentVec3 ( const sgLineSegment3 line,
+                                         const sgVec3 pnt )
+{
+  return sgSqrt ( sgDistSquaredToLineSegmentVec3(line,pnt) ) ;
+}
+
+
 inline SGfloat sgDistToPlaneVec3 ( const sgVec4 plane, const sgVec3 pnt )
 {
   return sgScalarProductVec3 ( plane, pnt ) + plane[3] ;
 }
- 
+
+
 inline SGfloat sgHeightAbovePlaneVec3 ( const sgVec4 plane, const sgVec3 pnt )
 {
-  return pnt[3] - sgHeightOfPlaneVec2 ( plane, pnt ) ;
+  return pnt[2] - sgHeightOfPlaneVec2 ( plane, pnt ) ;
 }
 
 extern void sgReflectInPlaneVec3 ( sgVec3 dst, const sgVec3 src, const sgVec4 plane ) ;
@@ -750,8 +884,25 @@ inline void sgMakePlane ( sgVec4 dst, const sgVec3 a, const sgVec3 b, const sgVe
   dst [ 3 ] = - sgScalarProductVec3 ( dst, a ) ;
 }
 
+float sgTriArea( sgVec3 p0, sgVec3 p1, sgVec3 p2 );
 
 
+// Fast code. Result is in the range  0..pi:
+inline SGfloat sgAngleBetweenNormalizedVec3 ( sgVec3 v1, sgVec3 v2 )
+{
+  float f = sgScalarProductVec3 ( v1, v2 ) ;
+  
+  return (float)(acos((f>=1.0f)?1.0f:(f<=-1.0f)?-1.0f:f)*SG_RADIANS_TO_DEGREES) ;
+}
+
+// Fast code. Result is in the range  0..pi:
+SGfloat sgAngleBetweenVec3 ( sgVec3 v1, sgVec3 v2 );
+
+// all three have to be normalized. Slow code. Result is in the range  0..2*pi:
+SGfloat sgAngleBetweenNormalizedVec3 (sgVec3 first, sgVec3 second, sgVec3 normal);
+
+// normal has to be normalized. Slow code. Result is in the range  0..2*pi:
+SGfloat sgAngleBetweenVec3 ( sgVec3 v1, sgVec3 v2, sgVec3 normal );
 
 class sgSphere
 {
@@ -778,7 +929,11 @@ public:
   int isEmpty (void) const { return radius <  SG_ZERO ; }
   void empty  (void)	   { radius = - SG_ONE ; }
 
-  void orthoXform ( const sgMat4 m ) { sgXformPnt3 ( center, center, m ) ; }
+  void orthoXform ( const sgMat4 m )
+  {
+    sgXformPnt3 ( center, center, m ) ;
+    // radius *= sgLengthVec3 ( m[0] ) ;  -- degrades performance for non-scaled matrices ...
+  }
 
   void extend ( const sgSphere *s ) ;
   void extend ( const sgBox    *b ) ;
@@ -902,7 +1057,9 @@ public:
     update () ;
   }
 
-  void setFrustum ( const SGfloat l, const SGfloat r, const SGfloat b, const SGfloat t, const SGfloat n, const SGfloat f )
+  void setFrustum ( const SGfloat l, const SGfloat r,
+                    const SGfloat b, const SGfloat t,
+                    const SGfloat n, const SGfloat f )
   {
     left  = l ; right = r ;
     top   = t ; bot   = b ;
@@ -1033,13 +1190,13 @@ void sgAngleAxisToQuat ( sgQuat dst,
 
 /* Convert a matrix to/from a quat */
 
-void sgMatrixToQuat ( sgQuat quat, sgMat4 m ) ;
-void sgQuatToMatrix ( sgMat4 m, sgQuat quat ) ;
+void sgMatrixToQuat ( sgQuat quat, const sgMat4 m ) ;
+void sgQuatToMatrix ( sgMat4 m, const sgQuat quat ) ;
 
 /* Convert a set of eulers to/from a quat */
 
-void sgQuatToEuler( SGfloat h, SGfloat p, SGfloat r, sgQuat quat ) ;
-void sgEulerToQuat( sgQuat quat, sgVec3 hpr ) ;
+void sgQuatToEuler( sgVec3 hpr, const sgQuat quat ) ;
+void sgEulerToQuat( sgQuat quat, const sgVec3 hpr ) ;
 
 inline void sgEulerToQuat( sgQuat dst,
                            SGfloat h, SGfloat p, SGfloat r )
@@ -1069,7 +1226,7 @@ inline void sgHPRToQuat ( sgQuat dst, const sgVec3 hpr )
   sgScaleVec3 ( tmp, hpr, SG_DEGREES_TO_RADIANS ) ;
 
   sgEulerToQuat ( dst, tmp ) ;
-};
+}
 
 /* Multiply quaternions together (concatenate rotations) */
 
@@ -1095,26 +1252,53 @@ inline void sgPreMultQuat ( sgQuat dst, const sgQuat q )
 /* Rotate a quaternion by a given angle and axis (convenience function) */
 
 
-inline void sgRotQuat ( sgQuat dst, const SGfloat angle, const sgVec3 axis )
+inline void sgPreRotQuat ( sgQuat dst, const SGfloat angle, const sgVec3 axis )
 {
   sgQuat q ;
+  sgAngleAxisToQuat ( q, angle, axis ) ;
+  sgPreMultQuat ( dst, q ) ;
+  sgNormaliseQuat ( dst ) ;
+}
 
+
+inline void sgPostRotQuat ( sgQuat dst, const SGfloat angle, const sgVec3 axis )
+{
+  sgQuat q ;
   sgAngleAxisToQuat ( q, angle, axis ) ;
   sgPostMultQuat ( dst, q ) ;
   sgNormaliseQuat ( dst ) ;
 }
 
 
-inline void sgRotQuat ( sgQuat dst,
+inline void sgPreRotQuat ( sgQuat dst,
                         const SGfloat angle,
                         const SGfloat x, const SGfloat y, const SGfloat z )
 {
   sgVec3 axis ;
 
   sgSetVec3 ( axis, x, y, z ) ;
-  sgRotQuat ( dst, angle, axis ) ;
+  sgPreRotQuat ( dst, angle, axis ) ;
 }
 
+inline void sgPostRotQuat ( sgQuat dst,
+                        const SGfloat angle,
+                        const SGfloat x, const SGfloat y, const SGfloat z )
+{
+  sgVec3 axis ;
+
+  sgSetVec3 ( axis, x, y, z ) ;
+  sgPostRotQuat ( dst, angle, axis ) ;
+}
+
+
+/* more meaningful names */
+#define sgWorldRotQuat  sgPostRotQuat
+#define sgLocalRotQuat  sgPreRotQuat
+
+/* for backwards compatibility */
+#define sgRotQuat       sgPostRotQuat
+
+
 /* SWC - Interpolate between to quaternions */
 
 extern void sgSlerpQuat ( sgQuat dst,
@@ -1138,8 +1322,7 @@ int sgIsectInfLineInfLine   ( sgVec3 dst,
 SGfloat sgIsectLinesegPlane ( sgVec3 dst,
                               sgVec3 v1, sgVec3 v2,
                               sgVec4 plane ) ;
-
-
+bool sgPointInTriangle      ( sgVec3 point, sgVec3 tri[3] );
 
 
 
@@ -1155,9 +1338,12 @@ SGfloat sgIsectLinesegPlane ( sgVec3 dst,
 #define SGD_TWO   2.0
 #define SGD_THREE 3.0
 #define SGD_45    45.0
+#define SGD_60    60.0
+#define SGD_90    90.0
 #define SGD_180   180.0
 #define SGD_MAX   DBL_MAX
 
+
 #define SGD_X	0
 #define SGD_Y	1
 #define SGD_Z	2
@@ -1181,6 +1367,21 @@ inline SGDfloat sgdSqrt   ( const SGDfloat x ) { return sqrt ( x ) ; }
 inline SGDfloat sgdSquare ( const SGDfloat x ) { return x * x ; }
 inline SGDfloat sgdAbs    ( const SGDfloat a ) { return ( a < SGD_ZERO ) ? -a : a ; }
 
+inline SGDfloat sgdASin ( SGDfloat s )
+                { return (SGDfloat) asin ( s ) * SGD_RADIANS_TO_DEGREES ; }
+inline SGDfloat sgdACos ( SGDfloat s )
+                { return (SGDfloat) acos ( s ) * SGD_RADIANS_TO_DEGREES ; }
+inline SGDfloat sgdATan ( SGDfloat s )
+                { return (SGDfloat) atan ( s ) * SGD_RADIANS_TO_DEGREES ; }
+inline SGDfloat sgdATan2 ( SGDfloat y, SGDfloat x )
+                { return (SGDfloat) atan2 ( y,x ) * SGD_RADIANS_TO_DEGREES ; }
+inline SGDfloat sgdSin ( SGDfloat s )
+                { return (SGDfloat) sin ( s * SGD_DEGREES_TO_RADIANS ) ; }
+inline SGDfloat sgdCos ( SGDfloat s )
+                { return (SGDfloat) cos ( s * SGD_DEGREES_TO_RADIANS ) ; }
+inline SGDfloat sgdTan ( SGDfloat s )
+                { return (SGDfloat) tan ( s * SGD_DEGREES_TO_RADIANS ) ; }
+
 inline int sgdCompareFloat ( const SGDfloat a, const SGDfloat b, const SGDfloat tol )
 {
   if ( ( a + tol ) < b ) return -1 ;
@@ -1312,6 +1513,7 @@ inline void sgdFullXformPnt4 ( sgdVec4 dst, const sgdMat4 mat ) { sgdFullXformPn
   ScalarProduct - scalar (dot) product.
   VectorProduct - vector (cross) product (3-element vectors ONLY!).
   Normalise/Normalize - make vector be one unit long.
+  Lerp      - linear interpolation by a fraction 'f'.
 */
 
 inline void sgdZeroVec2 ( sgdVec2 dst ) { dst[0]=dst[1]=SGD_ZERO ; }
@@ -1656,6 +1858,51 @@ inline SGDfloat sgdScalarProductVec4 ( const sgdVec4 a, const sgdVec4 b )
 
 extern void sgdVectorProductVec3 ( sgdVec3 dst, const sgdVec3 a, const sgdVec3 b ) ;
 
+
+inline void sgdLerpVec4 ( sgdVec4 dst, const sgdVec4 a, const sgdVec4 b, const SGDfloat f )
+{
+  dst[0] = a[0] + f * ( b[0] - a[0] ) ;
+  dst[1] = a[1] + f * ( b[1] - a[1] ) ;
+  dst[2] = a[2] + f * ( b[2] - a[2] ) ;
+  dst[3] = a[3] + f * ( b[3] - a[3] ) ;
+}
+
+inline void sgdLerpVec3 ( sgdVec3 dst, const sgdVec3 a, const sgdVec3 b, const SGDfloat f )
+{
+  dst[0] = a[0] + f * ( b[0] - a[0] ) ;
+  dst[1] = a[1] + f * ( b[1] - a[1] ) ;
+  dst[2] = a[2] + f * ( b[2] - a[2] ) ;
+}
+
+inline void sgdLerpVec2 ( sgdVec2 dst, const sgdVec2 a, const sgdVec2 b, const SGDfloat f )
+{
+  dst[0] = a[0] + f * ( b[0] - a[0] ) ;
+  dst[1] = a[1] + f * ( b[1] - a[1] ) ;
+}
+
+inline void sgdLerpAnglesVec3 ( sgdVec3 dst, const sgdVec3 a,
+                                             const sgdVec3 b,
+                                             const SGDfloat f )
+{
+  sgdVec3 tmp ;
+ 
+  if ( b[0] - a[0] >  180.0 ) tmp[0] = a[0] + 360.0 ; else
+  if ( b[0] - a[0] < -180.0 ) tmp[0] = a[0] - 360.0 ; else tmp[0] = a[0] ;
+ 
+  if ( b[1] - a[1] >  180.0 ) tmp[1] = a[1] + 360.0 ; else
+  if ( b[1] - a[1] < -180.0 ) tmp[1] = a[1] - 360.0 ; else tmp[1] = a[1] ;
+ 
+  if ( b[2] - a[2] >  180.0 ) tmp[2] = a[2] + 360.0 ; else
+  if ( b[2] - a[2] < -180.0 ) tmp[2] = a[2] - 360.0 ; else tmp[2] = a[2] ;
+ 
+  dst[0] = tmp[0] + f * ( b[0] - tmp[0] ) ;
+  dst[1] = tmp[1] + f * ( b[1] - tmp[1] ) ;
+  dst[2] = tmp[2] + f * ( b[2] - tmp[2] ) ;
+}                                                                               
+
+
+
+
 inline SGDfloat sgdDistanceSquaredVec2 ( const sgdVec2 a, const sgdVec2 b )
 {
   return sgdSquare ( a[0]-b[0] ) + sgdSquare ( a[1]-b[1] ) ;
@@ -1831,7 +2078,7 @@ inline SGDfloat sgdDistToPlaneVec3 ( const sgdVec4 plane, const sgdVec3 pnt )
  
 inline SGDfloat sgdHeightAbovePlaneVec3 ( const sgdVec4 plane, const sgdVec3 pnt )
 {
-  return pnt[3] - sgdHeightOfPlaneVec2 ( plane, pnt ) ;
+  return pnt[2] - sgdHeightOfPlaneVec2 ( plane, pnt ) ;
 }
 
 extern void sgdReflectInPlaneVec3 ( sgdVec3 dst, const sgdVec3 src, const sgdVec4 plane ) ;
@@ -1898,7 +2145,11 @@ public:
   int isEmpty (void) const { return radius <  SGD_ZERO ; }
   void empty  (void)       { radius = - SGD_ONE ; }
 
-  void orthoXform ( const sgdMat4 m ) { sgdXformPnt3 ( center, center, m ) ; }
+  void orthoXform ( const sgdMat4 m )
+  {
+    sgdXformPnt3 ( center, center, m ) ;
+    // radius *= sgdLengthVec3 ( m[0] ) ;
+  }
 
   void extend ( const sgdSphere *s ) ;
   void extend ( const sgdBox    *b ) ;
@@ -2021,7 +2272,9 @@ public:
     update () ;
   }
 
-  void setFrustum ( const SGDfloat l, const SGDfloat r, const SGDfloat b, const SGDfloat t, const SGDfloat n, const SGDfloat f )
+  void setFrustum ( const SGDfloat l, const SGDfloat r,
+                    const SGDfloat b, const SGDfloat t,
+                    const SGDfloat n, const SGDfloat f )
   {
     left  = l ; right = r ;
     top   = t ; bot   = b ;
@@ -2151,13 +2404,13 @@ void sgdAngleAxisToQuat ( sgdQuat dst,
 
 /* Convert a matrix to/from a quat */
 
-void sgdMatrixToQuat ( sgdQuat quat, sgdMat4 m ) ;
-void sgdQuatToMatrix ( sgdMat4 m, sgdQuat quat ) ;
+void sgdMatrixToQuat ( sgdQuat quat, const sgdMat4 m ) ;
+void sgdQuatToMatrix ( sgdMat4 m, const sgdQuat quat ) ;
 
 /* Convert a set of eulers to/from a quat */
 
-void sgdQuatToEuler( SGDfloat h, SGDfloat p, SGDfloat r, sgdQuat quat ) ;
-void sgdEulerToQuat( sgdQuat quat, sgdVec3 hpr ) ;
+void sgdQuatToEuler( sgdVec3 hpr, const sgdQuat quat ) ;
+void sgdEulerToQuat( sgdQuat quat, const sgdVec3 hpr ) ;
 
 inline void sgdEulerToQuat( sgdQuat dst,
                             SGDfloat h, SGDfloat p, SGDfloat r )
@@ -2187,7 +2440,7 @@ inline void sgdHPRToQuat ( sgdQuat dst, const sgdVec3 hpr )
   sgdScaleVec3 ( tmp, hpr, SGD_DEGREES_TO_RADIANS ) ;
 
   sgdEulerToQuat ( dst, tmp ) ;
-};
+}
 
 /* Multiply quaternions together (concatenate rotations) */
 
@@ -2329,6 +2582,304 @@ inline void sgdSetQuat ( sgdQuat dst, sgQuat src )
 }
 
 
+/*
+  Intersection testing.
+*/
+
+int sgdIsectPlanePlane        ( sgdVec3 point, sgdVec3 dir,
+				    sgdVec4 plane1, sgdVec4 plane2 ) ;
+int sgdIsectInfLinePlane      ( sgdVec3 dst,
+				    sgdVec3 l_org, sgdVec3 l_vec,
+				    sgdVec4 plane ) ;
+int sgdIsectInfLineInfLine    ( sgdVec3 dst,
+				    sgdVec3 l1_org, sgdVec3 l1_vec,
+				    sgdVec3 l2_org, sgdVec3 l2_vec ) ;
+SGDfloat sgdIsectLinesegPlane ( sgdVec3 dst,
+				    sgdVec3 v1, sgdVec3 v2,
+				    sgdVec4 plane ) ;
+bool sgdPointInTriangle       ( sgdVec3 point, sgdVec3 tri[3] );
+
+
+/*
+  TRIANGLE SOLVERS - These work for any triangle.
+
+  SSS  == Side-lengths for all three sides.
+  SAS  == Side-lengths for two sides - plus the angle between them.
+  ASA  == Two angles plus the length of the Side between them.
+  Area == The area of the triangle.
+*/
+
+SGfloat sgTriangleSolver_ASAtoArea ( SGfloat angA, SGfloat lenB, SGfloat angC );
+SGfloat sgTriangleSolver_SAStoArea ( SGfloat lenA, SGfloat angB, SGfloat lenC );
+SGfloat sgTriangleSolver_SSStoArea ( SGfloat lenA, SGfloat lenB, SGfloat lenC );
+SGfloat sgTriangleSolver_SAAtoArea ( SGfloat lenA, SGfloat angB, SGfloat angA );
+SGfloat sgTriangleSolver_ASStoArea ( SGfloat angB, SGfloat lenA, SGfloat lenB,
+                                     int angA_is_obtuse );
+
+void sgTriangleSolver_SSStoAAA ( SGfloat  lenA, SGfloat  lenB, SGfloat  lenC, 
+                                 SGfloat *angA, SGfloat *angB, SGfloat *angC ) ;
+void sgTriangleSolver_SAStoASA ( SGfloat  lenA, SGfloat  angB, SGfloat  lenC,
+                                 SGfloat *angA, SGfloat *lenB, SGfloat *angC ) ;
+void sgTriangleSolver_ASAtoSAS ( SGfloat  angA, SGfloat  lenB, SGfloat  angC,
+                                 SGfloat *lenA, SGfloat *angB, SGfloat *lenC ) ;
+void sgTriangleSolver_SAAtoASS ( SGfloat  lenA, SGfloat  angB, SGfloat  angA,
+                                 SGfloat *angC, SGfloat *lenB, SGfloat *lenC ) ;
+void sgTriangleSolver_ASStoSAA ( SGfloat  angB, SGfloat  lenA, SGfloat  lenB,
+                                 int angA_is_obtuse,
+                                 SGfloat *lenC, SGfloat *angA, SGfloat *angC ) ;
+
+
+SGDfloat sgdTriangleSolver_ASAtoArea(SGDfloat angA,SGDfloat lenB,SGDfloat angC);
+SGDfloat sgdTriangleSolver_SAStoArea(SGDfloat lenA,SGDfloat angB,SGDfloat lenC);
+SGDfloat sgdTriangleSolver_SSStoArea(SGDfloat lenA,SGDfloat lenB,SGDfloat lenC);
+SGDfloat sgdTriangleSolver_ASStoArea(SGDfloat angB,SGDfloat lenA,SGDfloat lenB);
+SGDfloat sgdTriangleSolver_SAAtoArea(SGDfloat lenA,SGDfloat angB,SGDfloat angA);
+
+void sgdTriangleSolver_SSStoAAA(SGDfloat  lenA,SGDfloat  lenB, SGDfloat  lenC, 
+                                SGDfloat *angA,SGDfloat *angB,SGDfloat *angC ) ;
+void sgdTriangleSolver_SAStoASA(SGDfloat  lenA,SGDfloat  angB,SGDfloat  lenC,
+                                SGDfloat *angA,SGDfloat *lenB,SGDfloat *angC ) ;
+void sgdTriangleSolver_ASAtoSAS(SGDfloat  angA,SGDfloat  lenB,SGDfloat  angC,
+                                SGDfloat *lenA,SGDfloat *angB,SGDfloat *lenC ) ;
+void sgdTriangleSolver_ASStoSAA(SGDfloat  angB,SGDfloat  lenA,SGDfloat  lenB,
+                                SGDfloat *lenC,SGDfloat *angA,SGDfloat *angC ) ;
+void sgdTriangleSolver_SAAtoASS(SGDfloat  lenA,SGDfloat  angB,SGDfloat  angA,
+                                SGDfloat *angC,SGDfloat *lenB,SGDfloat *lenC ) ;
+
+/*
+  SPRING-MASS-DAMPER (with simple Euler integrator)
+*/
+
+
+extern sgVec3 _sgGravity ;
+
+inline void   sgSetGravity     ( float  g ) { sgSetVec3 ( _sgGravity, 0.0f, 0.0f, -g ) ; }
+inline void   sgSetGravityVec3 ( sgVec3 g ) { sgCopyVec3 ( _sgGravity, g ) ; }
+inline float *sgGetGravityVec3 () { return   _sgGravity    ; }
+inline float  sgGetGravity     () { return - _sgGravity[2] ; }
+
+
+class sgParticle
+{
+  float ooMass ;  /* One-over-mass */
+  sgVec3 pos   ;
+  sgVec3 vel   ;
+  sgVec3 force ;
+
+public:
+
+  sgParticle ( float mass, sgVec3 _pos )
+  {
+    setMass    ( mass ) ;
+    sgCopyVec3 ( pos, _pos ) ;
+    sgZeroVec3 (  vel  ) ;
+    sgZeroVec3 ( force ) ;
+  }
+
+  sgParticle ( float mass, float x = 0.0f, float y = 0.0f, float z = 0.0f )
+  {
+    setMass    ( mass  ) ;
+    sgSetVec3  ( pos, x, y, z ) ;
+    sgZeroVec3 (  vel  ) ;
+    sgZeroVec3 ( force ) ;
+  }
+
+  float *getPos         () { return pos      ; }
+  float *getVel         () { return vel      ; }
+  float *getForce       () { return force    ; }
+  float  getOneOverMass () { return ooMass   ; }
+  float  getMass        () { return 1.0f / ooMass ; }
+
+  void   setPos   ( sgVec3 p ) { sgCopyVec3 ( pos  , p ) ; }
+  void   setVel   ( sgVec3 v ) { sgCopyVec3 ( vel  , v ) ; }
+  void   setForce ( sgVec3 f ) { sgCopyVec3 ( force, f ) ; }
+
+  void   setPos   ( float x, float y, float z ) { sgSetVec3 ( pos  ,x,y,z ) ; }
+  void   setVel   ( float x, float y, float z ) { sgSetVec3 ( vel  ,x,y,z ) ; }
+  void   setForce ( float x, float y, float z ) { sgSetVec3 ( force,x,y,z ) ; }
+
+  void   setOneOverMass ( float oom ) { ooMass = oom ; }
+
+  void   setMass ( float m )
+  {
+    assert ( m > 0.0f ) ;
+    ooMass = 1.0f / m ;
+  }
+
+  void zeroForce   ()           { sgZeroVec3   ( force ) ; }
+  void addForce    ( sgVec3 f ) { sgAddVec3    ( force, f ) ; }
+  void subForce    ( sgVec3 f ) { sgSubVec3    ( force, f ) ; }
+  void gravityOnly ()           { sgScaleVec3  ( force, sgGetGravityVec3 (), ooMass ) ; }
+
+  void bounce ( sgVec3 normal, float coefRestitution )
+  {
+    sgVec3 vn, vt ;
+    sgScaleVec3 ( vn, normal,
+                   sgScalarProductVec3 ( normal, vel ) ) ;
+    sgSubVec3 ( vt, vel, vn ) ;
+    sgAddScaledVec3 ( vel, vt, vn, -coefRestitution ) ;
+  }
+
+  void update ( float dt )
+  {
+    sgAddScaledVec3 ( vel, force, dt * ooMass ) ;
+    sgAddScaledVec3 ( pos, vel, dt ) ;
+  }
+} ;
+
+
+class sgSpringDamper
+{
+  sgParticle *p0 ;
+  sgParticle *p1 ;
+
+  float restLength ;
+  float stiffness  ;
+  float damping    ;
+
+public:
+
+  sgSpringDamper ()
+  {
+    p0 = p1 = NULL ;
+    stiffness  = 1.0f ;
+    damping    = 1.0f ;
+    restLength = 1.0f ;
+  }
+
+  sgSpringDamper ( sgParticle *_p0, sgParticle *_p1,
+                   float _stiffness, float _damping,
+                   float _restLength = -1.0f )
+  {
+    p0 = _p0 ;
+    p1 = _p1 ;
+    stiffness = _stiffness ;
+    damping   = _damping   ;
+
+    if ( _restLength < 0.0f )
+    {
+      if ( p0 != NULL && p1 != NULL )
+        restLength = sgDistanceVec3 ( p0->getPos(), p1->getPos() ) ;
+      else
+        restLength = _restLength ;
+    }
+    else
+      restLength = 1.0f ;
+  }
+
+  float       getRestLength () { return restLength ; }
+  float       getStiffness  () { return stiffness  ; }
+  float       getDamping    () { return damping    ; }
+
+  sgParticle *getParticle   ( int which ) { return ( which == 0 ) ? p0 : p1 ; }
+
+
+  void setParticles ( sgParticle *_p0, sgParticle *_p1 ) { p0 = _p0 ; p1 = _p1 ; }
+  void setParticle  ( int which, sgParticle *p ) { if ( which == 0 ) p0 = p ; else p1 = p ; }
+
+  void setRestLength () { restLength = sgDistanceVec3 ( p0->getPos(), p1->getPos() ) ; }
+
+  void setRestLength ( float l ) { restLength = l ; }
+  void setStiffness  ( float s ) { stiffness  = s ; }
+  void setDamping    ( float d ) { damping    = d ; }
+
+  void update ()
+  {
+    sgVec3 dP ; sgSubVec3 ( dP, p0->getPos(), p1->getPos() ) ;
+    sgVec3 dV ; sgSubVec3 ( dV, p0->getVel(), p1->getVel() ) ;
+
+    float  L = sgLengthVec3 ( dP ) ; if ( L == 0.0f ) L = 0.0000001f ;
+    float  H = ( L - restLength ) * stiffness ;
+    float  D = sgScalarProductVec3 ( dV, dP ) * damping / L ;
+
+    sgVec3 F ; sgScaleVec3 ( F, dP, - ( H + D ) / L ) ;
+
+    p0 -> addForce ( F ) ;
+    p1 -> subForce ( F ) ;
+  }
+
+} ;
+
+
+/*
+  It must be true that (x % NOISE_WRAP_INDEX) == (x & NOISE_MOD_MASK)
+  so NOISE_WRAP_INDEX must be a power of two, and NOISE_MOD_MASK must be
+  that power of 2 - 1.  as indices are implemented, as unsigned chars,
+  NOISE_WRAP_INDEX shoud be less than or equal to 256.
+  There's no good reason to change it from 256, really.
+
+  NOISE_LARGE_PWR2 is a large power of 2, we'll go for 4096, to add to
+  negative numbers in order to make them positive
+*/
+
+#define SG_PERLIN_NOISE_WRAP_INDEX    256
+#define SG_PERLIN_NOISE_MOD_MASK      255
+#define SG_PERLIN_NOISE_LARGE_PWR2   4096
+
+
+
+class sgPerlinNoise_1D
+{
+private:
+
+  SGfloat gradTable [ SG_PERLIN_NOISE_WRAP_INDEX * 2 + 2 ] ;
+
+public:
+
+  sgPerlinNoise_1D () ;
+
+  void regenerate () ;
+
+  SGfloat getNoise ( SGfloat x ) ;
+} ;
+
+
+
+class sgPerlinNoise_2D
+{
+private:
+
+  sgVec2 gradTable [ SG_PERLIN_NOISE_WRAP_INDEX * 2 + 2 ] ;
+
+public:
+
+  sgPerlinNoise_2D () ;
+
+  void regenerate () ;
+
+  SGfloat getNoise ( sgVec2 pos ) ;
+  SGfloat getNoise ( SGfloat x, SGfloat y )
+  {
+    sgVec2 p ;
+    sgSetVec2 ( p, x, y ) ;
+    return getNoise ( p ) ;
+  }
+} ;
+
+
+
+class sgPerlinNoise_3D
+{
+private:
+
+  sgVec3 gradTable [ SG_PERLIN_NOISE_WRAP_INDEX * 2 + 2 ] ;
+
+public:
+
+  sgPerlinNoise_3D () ;
+
+  void regenerate () ;
+
+  SGfloat getNoise ( sgVec3 pos ) ;
+  SGfloat getNoise ( SGfloat x, SGfloat y, SGfloat z )
+  {
+    sgVec3 p ;
+    sgSetVec3 ( p, x, y, z ) ;
+    return getNoise ( p ) ;
+  }
+} ;
+
+
 
 #endif
 
